<?php

namespace Drupal\parameters\Plugin\Parameter;

use Drupal\Component\Plugin\DependentPluginInterface;
use Drupal\Component\Utility\Html;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Entity\EntityRepositoryInterface;
use Drupal\Core\Entity\ContentEntityFormInterface;
use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Entity\Entity\EntityFormDisplay;
use Drupal\Core\Form\FormBuilderInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Form\SubformState;
use Drupal\Core\Form\SubformStateInterface;
use Drupal\Core\TypedData\DataDefinition;
use Drupal\Core\TypedData\TypedDataInterface;
use Drupal\Core\Utility\Token;
use Drupal\parameters\Entity\ParametersCollectionStorage;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\parameters\Normalizer\ParametersContentEntityNormalizer;
use Drupal\parameters\Normalizer\ParametersEntityReferenceFieldItemNormalizer;
use Drupal\parameters\Plugin\ParameterBase;
use Drupal\parameters\Plugin\ParameterInterface;
use Drupal\parameters\Plugin\ParameterManager;
use Drupal\parameters\Plugin\PropertyParameterInterface;
use Drupal\parameters\Plugin\UsageParameterInterface;
use Symfony\Component\Serializer\Serializer;
use Drupal\user\RoleInterface;

/**
 * Defines content as parameter.
 *
 * @Parameter(
 *   id = "content",
 *   label = @Translation("Content"),
 *   weight = 100,
 *   deriver = "Drupal\parameters\Plugin\Derivative\Parameter\ContentDeriver"
 * )
 */
class Content extends ParameterBase implements PropertyParameterInterface, UsageParameterInterface, DependentPluginInterface {

  /**
   * {@inheritdoc}
   */
  protected string $dataType = 'entity';

  /**
   * The entity form display to use for configuring the content entity.
   *
   * @var string
   */
  protected string $entityFormDisplay = 'parameters';

  /**
   * The configured content entity.
   *
   * @var \Drupal\Core\Entity\ContentEntityInterface
   */
  protected ContentEntityInterface $configuredContentEntity;

  /**
   * The form object used for configuring the content entity.
   *
   * @var \Drupal\Core\Entity\ContentEntityFormInterface|null
   */
  protected ?ContentEntityFormInterface $entityForm;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected EntityTypeManagerInterface $entityTypeManager;

  /**
   * The form builder.
   *
   * @var \Drupal\Core\Form\FormBuilderInterface
   */
  protected FormBuilderInterface $formBuilder;

  /**
   * The serializer.
   *
   * @var \Symfony\Component\Serializer\Serializer
   */
  protected Serializer $serializer;

  /**
   * The entity repository.
   *
   * @var \Drupal\Core\Entity\EntityRepositoryInterface
   */
  protected EntityRepositoryInterface $entityRepository;

  /**
   * The parameter plugin manager.
   *
   * @var \Drupal\parameters\Plugin\ParameterManager
   */
  protected ParameterManager $parameterManager;

  /**
   * The token service.
   *
   * @var \Drupal\Core\Utility\Token
   */
  protected Token $token;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    /** @var \Drupal\parameters\Plugin\Parameter\Content $instance */
    $instance = parent::create($container, $configuration, $plugin_id, $plugin_definition);
    $instance->setFormBuilder($container->get('form_builder'));
    $instance->setEntityTypeManager($container->get('entity_type.manager'));
    $instance->setSerializer($container->get('serializer'));
    $instance->setEntityRepository($container->get('entity.repository'));
    $instance->setParameterManager($container->get(ParameterManager::SERVICE_NAME));
    $instance->setToken($container->get('token'));

    if (isset($plugin_definition['entity_type'], $plugin_definition['bundle'])) {
      $entity_type_id = $plugin_definition['entity_type'];
      $bundle = $plugin_definition['bundle'];
    }
    else {
      [, $derivative_id] = explode(':', $plugin_id, 2);
      [$entity_type_id, $bundle] = explode('.', $derivative_id, 2);
    }
    $entity_type = $instance->entityTypeManager->getDefinition($entity_type_id);
    $instance->dataType = $entity_type->hasKey('bundle') ? 'entity:' . $entity_type_id . ':' . $bundle : 'entity:' . $entity_type_id;

    if (empty($instance->configuration['values'])) {
      $instance->configuration += $instance->defaultConfiguration();
    }
    $instance->initConfiguredContentEntity();
    return $instance;
  }

  /**
   * {@inheritdoc}
   */
  public function validate() {
    $plugin_definition = $this->getPluginDefinition();
    if (isset($plugin_definition['entity_type'], $plugin_definition['bundle'])) {
      $entity_type_id = $plugin_definition['entity_type'];
      $bundle = $plugin_definition['bundle'];
    }
    else {
      [$entity_type_id, $bundle] = explode('.', $this->getDerivativeId(), 2);
    }
    $entity_type = $this->entityTypeManager->getDefinition($entity_type_id);
    if ($entity_type->hasKey('uuid') && !isset($this->configuration['values'][$entity_type->getKey('uuid')])) {
      return $this->t('A uuid must be set.');
    }
    if ($entity_type->hasKey('bundle') && !isset($this->configuration['values'][$entity_type->getKey('bundle')])) {
      return $this->t('The bundle must be set.');
    }
    return parent::validate();
  }

  /**
   * {@inheritdoc}
   */
  public function getProcessedData(): TypedDataInterface {
    if (!isset($this->processedData)) {
      $this->processedData = $this->typedDataManager->createInstance($this->dataType, [
        'data_definition' => DataDefinition::createFromDataType($this->dataType),
        'name' => NULL,
        'parent' => NULL,
      ]);
      $this->processedData->setValue($this->processDataValue());
    }
    return $this->processedData;
  }

  /**
   * Processes the data value to be returned via ::getProcessedData().
   *
   * @return mixed
   *   The data value. This value must be compatible with the typed data type
   *   that is defined via ::$dataType, because it will be set as value for
   *   an instance of that typed data type.
   */
  protected function processDataValue() {
    $entity = $this->getConfiguredContentEntity();

    // Save the entity, if not yet available.
    if ($entity->isNew()) {
      $entity = $this->saveEntity($entity);
    }

    return $entity;
  }

  /**
   * {@inheritdoc}
   */
  public function getPreview(): string {
    $entity = $this->getConfiguredContentEntity();
    $label = (string) Html::escape($entity->label());
    return !$entity->isNew() && $entity->hasLinkTemplate('canonical') && $entity->access('view') ? (string) $entity->toLink($label, 'canonical')->toString() : $label;
  }

  /**
   * Set the entity type manager.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   */
  public function setEntityTypeManager(EntityTypeManagerInterface $entity_type_manager): void {
    $this->entityTypeManager = $entity_type_manager;
  }

  /**
   * Get the entity type manager.
   *
   * @return \Drupal\Core\Entity\EntityTypeManagerInterface
   *   The entity type manager.
   */
  public function getEntityTypeManager(): EntityTypeManagerInterface {
    if (!isset($this->entityTypeManager)) {
      $this->entityTypeManager = \Drupal::service('entity_type.manager');
    }
    return $this->entityTypeManager;
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    if ($plugin_definition = $this->getPluginDefinition()) {
      $values = [];

      if (isset($plugin_definition['entity_type'], $plugin_definition['bundle'])) {
        $entity_type_id = $plugin_definition['entity_type'];
        $bundle = $plugin_definition['bundle'];
      }
      else {
        [$entity_type_id, $bundle] = explode('.', $this->getDerivativeId(), 2);
      }

      $entity_type = $this->getEntityTypeManager()->getDefinition($entity_type_id);
      if ($entity_type->hasKey('bundle')) {
        $values[$entity_type->getKey('bundle')] = $bundle;
      }

      return ['values' => $values];
    }
    return [];
  }

  /**
   * Returns a JSON-formatted string representation of the given entity.
   *
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity
   *   The entity to format as JSON.
   *
   * @return string
   *   The JSON string.
   */
  public function toJson(ContentEntityInterface $entity): string {
    return $this->getSerializer()->serialize($entity, 'json', ['json_encode_options' => JSON_PRETTY_PRINT | JSON_UNESCAPED_UNICODE | JSON_UNESCAPED_SLASHES]);
  }

  /**
   * Converts the given JSON-formatted string to an entity object.
   *
   * @param string $json
   *   The JSON-formatted string.
   * @param string $entity_class
   *   The entity class to expect.
   *
   * @return \Drupal\Core\Entity\ContentEntityInterface
   *   The entity.
   */
  public function fromJson(string $json, string $entity_class): ContentEntityInterface {
    return $this->getSerializer()->deserialize($json, $entity_class, 'json');
  }

  /**
   * Get the serializer.
   *
   * @return \Symfony\Component\Serializer\Serializer
   *   The serializer.
   */
  public function getSerializer(): Serializer {
    if (!isset($this->serializer)) {
      $this->serializer = \Drupal::service('serializer');
    }
    return $this->serializer;
  }

  /**
   * Set the serializer.
   *
   * @param \Symfony\Component\Serializer\Serializer $serializer
   *   The serializer.
   */
  public function setSerializer(Serializer $serializer): void {
    $this->serializer = $serializer;
  }

  /**
   * Returns a normalized array of the given entity, suitable for configuration.
   *
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity
   *   The entity to normalize.
   *
   * @return array
   *   The normalized array.
   */
  public function toConfigArray(ContentEntityInterface $entity): array {
    ParametersContentEntityNormalizer::$cleanupFieldValues = TRUE;
    ParametersEntityReferenceFieldItemNormalizer::$normalizeNewEntities = TRUE;
    try {
      return $this->getSerializer()->normalize($entity, get_class($entity));
    }
    finally {
      ParametersContentEntityNormalizer::$cleanupFieldValues = FALSE;
      ParametersEntityReferenceFieldItemNormalizer::$normalizeNewEntities = FALSE;
    }
  }

  /**
   * Converts the given normalized config array to an entity object.
   *
   * @param array $array
   *   The normalized array.
   * @param string $entity_class
   *   The entity class to expect.
   *
   * @return \Drupal\Core\Entity\ContentEntityInterface
   *   The entity.
   */
  public function fromConfigArray(array $array, string $entity_class): ContentEntityInterface {
    ParametersContentEntityNormalizer::$cleanupFieldValues = TRUE;
    ParametersEntityReferenceFieldItemNormalizer::$normalizeNewEntities = TRUE;
    try {
      return $this->getSerializer()->denormalize($array, $entity_class);
    }
    finally {
      ParametersContentEntityNormalizer::$cleanupFieldValues = FALSE;
      ParametersEntityReferenceFieldItemNormalizer::$normalizeNewEntities = FALSE;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    $form['entity_form_info'] = [
      '#type' => 'container',
      'display_mode' => [
        '#markup' => $this->t('This configuration is using the "@mode" form display mode. <a href=":url" target="_blank">Manage form display modes</a>.', [
          '@mode' => $this->entityFormDisplay,
          ':url' => '/admin/structure/display-modes/form',
        ]),
        '#weight' => 10,
      ],
      '#weight' => -50,
    ];
    // We need to use a process callback for embedding the entity fields,
    // because the fields to embed need to know their "#parents".
    $wrapper_id = Html::getUniqueId('entity-content-fields');
    $entity = $this->getConfiguredContentEntity();
    $form['values'] = [
      '#prefix' => '<div id="' . $wrapper_id . '">',
      '#suffix' => '</div>',
      '#wrapper_id' => $wrapper_id,
      '#parameters__entity' => $entity,
      '#parameters__form_display' => $this->entityFormDisplay,
      '#process' => [[static::class, 'processContentConfigurationForm']],
    ];
    return $form;
  }

  /**
   * Process callback to insert the content entity form.
   *
   * @param array $element
   *   The containing element.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state.
   *
   * @return array
   *   The containing element, with the content entity form inserted.
   */
  public static function processContentConfigurationForm(array $element, FormStateInterface $form_state) {
    $entity = $element['#parameters__entity'];
    $form_display_mode = $element['#parameters__form_display'];
    $wrapper_id = $element['#wrapper_id'];
    $form_display = EntityFormDisplay::collectRenderDisplay($entity, $form_display_mode);

    // A very special treatment for the moderation state field that is coming
    // from core's content_moderation module. We need to wrap the according
    // widget with a decorator that prevents that widget from manipulating
    // entity values. This happens when the configured moderation state differs
    // from the initial state.
    if ($form_display->getComponent('moderation_state')) {
      $closure = \Closure::fromCallable(function () {
        /** @var \Drupal\Core\Entity\Entity\EntityFormDisplay $this */
        $this->getRenderer('moderation_state');
        $workaround_class = 'Drupal\flow\Workaround\ModerationStateWidgetWorkaround';
        $this->plugins['moderation_state'] = new $workaround_class($this->plugins['moderation_state']);
      });
      $closure->call($form_display);
    }

    $content_config_entities = $form_state->get('parameters__content_configuration') ?? [];
    $content_config_entities[$wrapper_id] = [$entity, $form_display];
    $form_state->set('parameters__content_configuration', $content_config_entities);
    $form_display->buildForm($entity, $element, $form_state);
    if ($entity->getEntityTypeId() === 'user') {
      self::processUserAccountForm($element, $form_state);
    }
    return $element;
  }

  /**
   * {@inheritdoc}
   */
  public function validateConfigurationForm(array &$form, FormStateInterface $form_state) {
    if (!$form_state->hasValue(['values']) || !isset($form['values']['#wrapper_id'])) {
      return;
    }
    $wrapper_id = $form['values']['#wrapper_id'];
    $content_config_entities = $form_state->get('parameters__content_configuration') ?? [];
    if (!isset($content_config_entities[$wrapper_id])) {
      return;
    }

    $complete_form_state = $form_state instanceof SubformStateInterface ? $form_state->getCompleteFormState() : $form_state;
    [$entity, $form_display] = $content_config_entities[$wrapper_id];
    $extracted = $form_display->extractFormValues($entity, $form['values'], $complete_form_state);
    // Extract the values of fields that are not rendered through widgets, by
    // simply copying from top-level form values. This leaves the fields that
    // are not being edited within this form untouched.
    // @see \Drupal\Tests\field\Functional\NestedFormTest::testNestedEntityFormEntityLevelValidation()
    foreach ($form_state->getValue(['values']) as $name => $values) {
      if ($entity->hasField($name) && !isset($extracted[$name])) {
        $entity->set($name, $values);
      }
    }
    $form_display->validateFormValues($entity, $form['values'], $complete_form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
    if (!$form_state->hasValue(['values']) || !isset($form['values']['#wrapper_id'])) {
      return;
    }
    $wrapper_id = $form['values']['#wrapper_id'];
    $content_config_entities = $form_state->get('parameters__content_configuration') ?? [];
    if (!isset($content_config_entities[$wrapper_id])) {
      return;
    }

    /** @var \Drupal\Core\Entity\ContentEntityInterface $entity */
    [$entity] = $content_config_entities[$wrapper_id];

    $form_display = EntityFormDisplay::collectRenderDisplay($entity, $this->entityFormDisplay, TRUE);
    $complete_form_state = $form_state instanceof SubformStateInterface ? $form_state->getCompleteFormState() : $form_state;
    $extracted = $form_display->extractFormValues($entity, $form['values'], $complete_form_state);
    foreach ($form_state->getValue(['values']) as $name => $values) {
      if ($entity->hasField($name) && !isset($extracted[$name])) {
        $entity->set($name, $values);
      }
    }

    // Filter field values that are not available on the form display mode.
    $entity_type = $entity->getEntityType();
    $entity_keys = $entity_type->getKeys();
    $components = $form_display->getComponents();
    foreach (array_keys($values) as $k_1) {
      if (!isset($components[$k_1]) && !in_array($k_1, $entity_keys)) {
        unset($values[$k_1]);
      }
    }

    if ($entity->getEntityTypeId() === 'user') {
      self::submitUserAccountForm($form['values'], SubformState::createForSubform($form['values'], $form, $form_state));
    }

    $this->setConfiguredContentEntity($entity);
    $values = $this->toConfigArray($entity);

    $this->configuration['values'] = $values;
  }

  /**
   * Instantiates the configured content entity.
   *
   * @return \Drupal\Core\Entity\ContentEntityInterface
   *   The initialized entity.
   */
  public function initConfiguredContentEntity(): ContentEntityInterface {
    if (isset($plugin_definition['entity_type'], $plugin_definition['bundle'])) {
      $entity_type_id = $plugin_definition['entity_type'];
      $bundle = $plugin_definition['bundle'];
    }
    else {
      [$entity_type_id, $bundle] = explode('.', $this->getDerivativeId(), 2);
    }
    $entity_type = $this->entityTypeManager->getDefinition($entity_type_id);

    $values = $this->configuration['values'];
    $uuid = $entity_type->hasKey('uuid') && isset($values[$entity_type->getKey('uuid')]) ? $values[$entity_type->getKey('uuid')] : NULL;
    while (is_array($uuid)) {
      $uuid = reset($uuid);
    }
    if ($uuid && ($entity = $this->entityRepository->loadEntityByUuid($entity_type_id, $uuid))) {
      $this->setConfiguredContentEntity($entity);
    }
    else {
      $this->setConfiguredContentEntity($this->fromConfigArray($values, $entity_type->getClass()));
    }

    return $this->configuredContentEntity;
  }

  /**
   * Get the configured content entity object.
   *
   * @return \Drupal\Core\Entity\ContentEntityInterface
   *   The initialized entity.
   */
  public function getConfiguredContentEntity(): ContentEntityInterface {
    if (!isset($this->configuredContentEntity)) {
      $this->initConfiguredContentEntity();
    }
    return $this->configuredContentEntity;
  }

  /**
   * Set the configured content entity.
   *
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity
   *   The entity to set as configured content entity.
   */
  public function setConfiguredContentEntity(ContentEntityInterface $entity): void {
    $this->configuredContentEntity = $entity;
  }

  /**
   * Process callback to insert elements of the user account form.
   *
   * @param array &$form
   *   The form element being processed.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state.
   */
  public static function processUserAccountForm(array &$form, FormStateInterface $form_state): void {
    /** @var \Drupal\user\UserInterface $account */
    $account = $form['#parameters__entity'];

    $form['status'] = [
      '#type' => 'radios',
      '#title' => t('Status'),
      '#default_value' => (int) $account->get('status')->value,
      '#options' => [t('Blocked'), t('Active')],
    ];

    $roles = array_map(['\Drupal\Component\Utility\Html', 'escape'], user_role_names(TRUE));

    $form['roles'] = [
      '#type' => 'checkboxes',
      '#title' => t('Roles'),
      '#default_value' => $account->getRoles(),
      '#options' => $roles,
    ];

    // Special handling for the inevitable "Authenticated user" role.
    $form['roles'][RoleInterface::AUTHENTICATED_ID] = [
      '#default_value' => TRUE,
      '#access' => FALSE,
    ];
  }

  /**
   * Submit callback for elements of the user account form.
   *
   * @param array &$element
   *   The element being processed.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state.
   */
  public static function submitUserAccountForm(array &$form, FormStateInterface $form_state): void {
    $wrapper_id = $form['#wrapper_id'];
    $content_config_entities = $form_state->get('parameters__content_configuration') ?? [];
    if (!isset($content_config_entities[$wrapper_id])) {
      return;
    }

    /** @var \Drupal\user\UserInterface $account */
    [$account] = $content_config_entities[$wrapper_id];

    if ($form_state->hasValue('roles') && is_string(key($form_state->getValue('roles')))) {
      $form_state->setValue('roles', array_keys(array_filter($form_state->getValue('roles'))));
    }

    $roles = [];
    foreach ($form_state->getValue('roles', []) as $rk => $rv) {
      if ($rv && $rv !== RoleInterface::AUTHENTICATED_ID) {
        $roles[] = $rv;
      }
    }
    $account->get('roles')->setValue($roles);
  }

  /**
   * Set the form builder.
   *
   * @param \Drupal\Core\Form\FormBuilderInterface $form_builder
   *   The form builder.
   */
  public function setFormBuilder(FormBuilderInterface $form_builder): void {
    $this->formBuilder = $form_builder;
  }

  /**
   * Get the form builder.
   *
   * @return \Drupal\Core\Form\FormBuilderInterface
   *   The form builder.
   */
  public function getFormBuilder(): FormBuilderInterface {
    if (!isset($this->formBuilder)) {
      $this->formBuilder = \Drupal::service('form_builder');
    }
    return $this->formBuilder;
  }

  /**
   * Set the entity repository.
   *
   * @param \Drupal\Core\Entity\EntityRepositoryInterface $entity_repository
   *   The entity repository.
   */
  public function setEntityRepository(EntityRepositoryInterface $entity_repository): void {
    $this->entityRepository = $entity_repository;
  }

  /**
   * Get the entity repository.
   *
   * @return \Drupal\Core\Entity\EntityRepositoryInterface
   *   The entity repository.
   */
  public function getEntityRepository(): EntityRepositoryInterface {
    if (!isset($this->entityRepository)) {
      $this->entityRepository = \Drupal::service('entity.repository');
    }
    return $this->entityRepository;
  }

  /**
   * Set the parameter plugin manager.
   *
   * @param \Drupal\parameters\Plugin\ParameterManager $manager
   *   The parameter plugin manager.
   */
  public function setParameterManager(ParameterManager $manager) {
    $this->parameterManager = $manager;
  }

  /**
   * Get the token service.
   *
   * @return \Drupal\Core\Utility\Token
   *   The token service.
   */
  public function getToken(): Token {
    if (!isset($this->token)) {
      $this->token = \Drupal::token();
    }
    return $this->token;
  }

  /**
   * Set the token service.
   *
   * @param \Drupal\Core\Utility\Token $token
   *   The token service.
   */
  public function setToken(Token $token): void {
    $this->token = $token;
  }

  /**
   * {@inheritdoc}
   */
  public function calculateDependencies() {
    $dependencies = [];

    $plugin_definition = $this->getPluginDefinition();
    if (isset($plugin_definition['entity_type'], $plugin_definition['bundle'])) {
      $entity_type_id = $plugin_definition['entity_type'];
      $bundle = $plugin_definition['bundle'];
    }
    else {
      [$entity_type_id, $bundle] = explode('.', $this->getDerivativeId(), 2);
    }
    $entity_type = $this->entityTypeManager->getDefinition($entity_type_id);
    if (!in_array($entity_type->getProvider(), ['core', 'component'])) {
      $dependencies['module'][] = $entity_type->getProvider();
    }
    if ($bundle_entity_type_id = $entity_type->getBundleEntityType()) {
      if ($bundle_config = $this->entityTypeManager->getStorage($bundle_entity_type_id)->load($bundle)) {
        $dependencies[$bundle_config->getConfigDependencyKey()][] = $bundle_config->getConfigDependencyName();
      }
    }

    $target_uuids = [];
    foreach (($this->configuration['values'] ?? []) as $field) {
      if (is_array($field)) {
        foreach ($field as $value) {
          if (is_array($value) && isset($value['target_uuid'])) {
            $target_uuids[$value['target_uuid']] = $value['target_uuid'];
          }
        }
      }
    }
    if ($target_uuids) {
      /** @var \Drupal\parameters\Entity\ParametersCollectionInterface $collection */
      foreach ($this->entityTypeManager->getStorage('parameters_collection')->loadMultiple() as $collection) {
        foreach (($collection->get('parameters') ?? []) as $parameter_config) {
          if (isset($parameter_config['values']['uuid'][0]['value']) && in_array($parameter_config['values']['uuid'][0]['value'], $target_uuids, TRUE)) {
            if (!isset(ParametersCollectionStorage::$saveStack[$collection->id()])) {
              $dependencies[$collection->getConfigDependencyKey()][] = $collection->getConfigDependencyName();
              continue 2;
            }
          }
        }
      }
    }

    return $dependencies;
  }

  /**
   * {@inheritdoc}
   */
  public function getProperty(string $name): ?ParameterInterface {
    $entity = $this->getProcessedData()->getValue();
    if (!($entity instanceof ContentEntityInterface)) {
      return NULL;
    }

    // Use the Token replacement system to retrieve a string value. For that,
    // the token type needs to be determined first.
    if (\Drupal::hasService('token.entity_mapper')) {
      /** @var \Drupal\token\TokenEntityMapperInterface $token_entity_mapper */
      $token_entity_mapper = \Drupal::service('token.entity_mapper');
      $token_type = $token_entity_mapper->getTokenTypeForEntityType($entity->getEntityTypeId(), TRUE);
    }
    else {
      $token_type = $entity->getEntityType()->get('token_type') ?: $entity->getEntityTypeId();
    }

    // Now build up the token and replace it with a value.
    if (strpos($name, '.')) {
      $name = str_replace('.', ':', $name);
    }
    $token = "[{$token_type}:{$name}]";
    $value = (string) $this->getToken()->replace($token, [$token_type => $entity], [
      'clear' => TRUE,
      'langcode' => $entity->language()->getId(),
    ]);
    if ($value !== '') {
      return $this->parameterManager->createInstance('string', ['value' => $value]);
    }

    // Last chance to get a field value.
    if ($entity->hasField($name) && !$entity->get($name)->isEmpty()) {
      return $this->parameterManager->createInstance('string', [
        'value' => $entity->get($name)->getString(),
      ]);
    }

    return NULL;
  }

  /**
   * {@inheritdoc}
   */
  public function onAddition(): void {
    $this->synchronizeContent();
  }

  /**
   * {@inheritdoc}
   */
  public function onChange(array $previous_configuration): void {
    $this->synchronizeContent();
  }

  /**
   * {@inheritdoc}
   */
  public function onRemoval(): void {
    // Do nothing. Leave the content entity as is, if it exists.
  }

  /**
   * Synchronizes the content entity with configured values.
   */
  public function synchronizeContent(): void {
    $plugin_definition = $this->getPluginDefinition();
    if (isset($plugin_definition['entity_type'], $plugin_definition['bundle'])) {
      $entity_type_id = $plugin_definition['entity_type'];
      $bundle = $plugin_definition['bundle'];
    }
    else {
      [$entity_type_id, $bundle] = explode('.', $this->getDerivativeId(), 2);
    }
    $entity_type = $this->entityTypeManager->getDefinition($entity_type_id);
    $id_fields = [$entity_type->getKey('id')];
    if ($entity_type->hasKey('revision')) {
      $id_fields[] = $entity_type->getKey('revision');
    }

    $configured = $this->fromConfigArray($this->configuration['values'], $entity_type->getClass());
    $loaded = $this->getConfiguredContentEntity();
    foreach ($loaded->getFields() as $field_name => $items) {
      if (in_array($field_name, $id_fields, TRUE)) {
        continue;
      }
      $items->setValue($configured->get($field_name)->getValue());
    }
    $entity = $this->saveEntity($loaded);
    $this->setConfiguredContentEntity($entity);
  }

  /**
   * Saves the entity.
   *
   * @param \Drupal\Core\Entity\ContentEntityInterface $entity
   *   The entity to save.
   *
   * @return |Drupal\Core\Entity\ContentEntityInterface
   *   The saved entity.
   */
  protected function saveEntity(ContentEntityInterface $entity): ContentEntityInterface {
    $lock_name = $entity->getEntityType()->hasKey('uuid') ? 'parameters:content:' . $entity->uuid() : 'parameters:content:' . ($entity->id() ?? uniqid());
    if (\Drupal::lock()->acquire($lock_name)) {
      $entity->save();
      \Drupal::lock()->release($lock_name);
    }
    elseif ($entity->uuid()) {
      // Someone else is already saving it. Wait a bit and then try to get it.
      $i = 0;
      $loaded = NULL;
      while (!($loaded = $this->entityRepository->loadEntityByUuid($entity->getEntityTypeId(), $entity->uuid()))) {
        $i++;
        usleep(10000);
        if ($i > 1000) {
          // More than 10 seconds have elapsed without any luck, abort.
          \Drupal::logger('parameters')->error(sprintf("Failed to load an entity from repository. Type: %s, UUID: %s", $entity->getEntityTypeId(), $entity->uuid()));
          break;
        }
      }
      $entity = $loaded ?? $entity;
    }
    return $entity;
  }

}
