<?php

namespace Drupal\parameters\Plugin\Parameter;

use Drupal\Component\Plugin\DependentPluginInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Render\RendererInterface;
use Drupal\parameters\Plugin\ParameterBase;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Defines formatted text as parameter.
 *
 * @Parameter(
 *   id = "text",
 *   label = @Translation("Formatted text")
 * )
 */
class Text extends ParameterBase implements DependentPluginInterface {

  /**
   * The module handler.
   *
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  protected ModuleHandlerInterface $moduleHandler;

  /**
   * The renderer.
   *
   * @var \Drupal\Core\Render\RendererInterface
   */
  protected RendererInterface $renderer;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    /** @var \Drupal\parameters\Plugin\Parameter\FormattedText $instance */
    $instance = parent::create($container, $configuration, $plugin_id, $plugin_definition);
    $instance->setModuleHandler($container->get('module_handler'));
    $instance->setRenderer($container->get('renderer'));
    return $instance;
  }

  /**
   * {@inheritdoc}
   */
  public function validateConfigurationForm(array &$form, FormStateInterface $form_state) {}

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    $form = parent::buildConfigurationForm($form, $form_state);
    if ($this->moduleHandler->moduleExists('filter')) {
      $form['value']['#type'] = 'text_format';
      $form['value']['#format'] = $this->configuration['format'];
    }
    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
    $value = $form_state->getValue('value');
    if (is_string($value)) {
      $this->configuration['value'] = $value;
      $this->configuration['format'] = NULL;
    }
    else {
      $this->configuration['value'] = $value['value'];
      $this->configuration['format'] = $value['format'];
    }
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return [
      'value' => '',
      'format' => function_exists('filter_default_format') ? filter_default_format() : NULL,
    ];
  }

  /**
   * Set the module handler.
   *
   * @param \Drupal\Core\Extension\ModuleHandlerInterface $module_handler
   *   The module handler.
   */
  public function setModuleHandler(ModuleHandlerInterface $module_handler): void {
    $this->moduleHandler = $module_handler;
  }

  /**
   * Set the renderer.
   *
   * @param \Drupal\Core\Render\RendererInterface $renderer
   *   The renderer.
   */
  public function setRenderer(RendererInterface $renderer) {
    $this->renderer = $renderer;
  }

  /**
   * {@inheritdoc}
   */
  public function calculateDependencies() {
    $dependencies = [];
    if ($this->moduleHandler->moduleExists('filter')) {
      $dependencies['module'][] = 'filter';
    }
    if (!empty($this->configuration['format']) && ($format = \Drupal::entityTypeManager()->getStorage('filter_format')->load($this->configuration['format']))) {
      $dependencies[$format->getConfigDependencyKey()][] = $format->getConfigDependencyName();
    }
    return $dependencies;
  }

  /**
   * {@inheritdoc}
   */
  protected function processDataValue() {
    $value = $this->configuration['value'];
    $format = $this->configuration['format'];
    if (!isset($format) || !$this->moduleHandler->moduleExists('filter')) {
      $build = ['#markup' => $value];
    }
    else {
      $build = [
        '#type' => 'processed_text',
        '#text' => $value,
        '#format' => $format,
      ];
    }
    return trim((string) $this->renderer->renderPlain($build));
  }

}
